from pyrecdp.primitives.generators import *
from .BasePipeline import BasePipeline
import logging
from pyrecdp.core.dataframe import DataFrameAPI
from pyrecdp.core import SeriesSchema
import pandas as pd
import copy

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class FeatureWrangler(BasePipeline):
    def __init__(self, dataset, label, supplementary_datasets = None, *args, **kwargs):
        super().__init__(dataset, label)
        self.data_profiler = [cls() for cls in feature_infer_list]
        self.pre_feature = [cls() for cls in label_feature_generator_list]
        # If we provided multiple datasets in this workload
        self.generators.append([cls() for cls in pre_feature_generator_list])
        self.generators.append([cls() for cls in transformation_generator_list])
        self.generators.append([cls() for cls in pre_enocode_feature_generator_list])
        self.generators.append([cls() for cls in local_encode_generator_list])
        self.generators.append([cls() for cls in global_dict_index_generator_list])
        self.generators.append([cls() for cls in post_feature_generator_list])
        self.generators.append([cls(final = True) for cls in final_generator_list])

        self.fit_analyze()

    def fit_analyze(self, *args, **kwargs):
        X = DataFrameAPI().instiate(self.dataset[self.main_table])
        sampled_data = X.may_sample()
        child = list(self.pipeline.keys())[-1]
        max_id = child
        if self.y is not None:
            # insert label process to pipeline            
            cur_id = child
            original_first_op = copy.deepcopy(self.pipeline[cur_id])
            self.pipeline[cur_id].output = [SeriesSchema(sampled_data[self.y])]
            
            for generator in self.data_profiler:
                self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id, pd.DataFrame(sampled_data[self.y]))
            for generator in self.pre_feature:
                self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id)
            
            cur_id = max_id + 1
            original_first_op.idx = cur_id
            original_first_op.op = "DataFrame"
            original_first_op.children = [child]
            self.y = [i.name for i in self.pipeline[child].output][0]
            self.pipeline[cur_id] = original_first_op            
            max_id = cur_id
            child = cur_id
            sampled_data = sampled_data[self.feature_columns]
        
        for generator in self.data_profiler:
            self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id, sampled_data)
        super().fit_analyze(*args, **kwargs)


from .base import BaseOperation
from .featuretools_adaptor import FeaturetoolsOperation
import copy
from pyrecdp.core.utils import class_name_fix

class HaversineOperation(FeaturetoolsOperation):
    def __init__(self, op_base):
        super().__init__(op_base)

    def get_function_pd(self):
        feature_in_out_map = copy.deepcopy(self.feature_in_out_map)
        def generate_ft_feature(df):
            for inputs_str, op in feature_in_out_map.items():
                inputs = eval(inputs_str)
                op_object = class_name_fix(op[1])()
                df[op[0]] = op_object(df[inputs[0]], df[inputs[1]])
            return df
        return generate_ft_feature




from .base import BaseFeatureGenerator as super_class
from pyrecdp.primitives.operations import Operation
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD
 
class DropUselessFeatureGenerator(super_class):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)
        self._astype_feature_map = None
        self.feature_in = []
        self.final = final

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if not self.final:
                if not (pa_field.is_numeric or pa_field.is_categorical):
                    self.feature_in.append(pa_field.name)
                    is_useful = True
            else:
                if not (pa_field.is_numeric):
                    self.feature_in.append(pa_field.name)
                    is_useful = True
        ret_schema = []
        for pa_field in pa_schema:
            if pa_field.name not in self.feature_in:
                ret_schema.append(pa_field)
        if is_useful:
            cur_idx = max_idx + 1
            config = self.feature_in
            pipeline[cur_idx] = Operation(cur_idx, children, ret_schema, op = 'drop', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx


from .base_api import base_api

class download(base_api):
    def __init__(self, name, url, unzip = False):
        super().__init__()      
        self.saved_path = self.download_url(name, url, unzip = unzip)
        print(f"Data is downloaded, use {self.saved_path} to open")
         


from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
import copy
import numpy as np
from IPython.display import display

class OnehotEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        config = copy.deepcopy(self.config)
        def encode(df):
            for feature, keys in config.items():
                selected_columns = [f"{feature}__{key}" for key in keys]
                one_hot_df = pd.get_dummies(df[feature], prefix = f"{feature}_")
                one_hot_df = one_hot_df.loc[:, one_hot_df.columns.isin(selected_columns)]
                #there is possibility designed keys are not included or new keys added
                # ret = {}
                # len_df = one_hot_df.shape[0]
                # for col_name in selected_columns:
                #     if col_name in one_hot_df.columns:
                #         ret[col_name] = one_hot_df[col_name].to_numpy()
                #     else:
                #         ret[col_name] = np.zeros(len_df, dtype=np.uint8)
                # one_hot_df = pd.DataFrame(ret, dtype=np.uint8, index=df.index)
                df = pd.concat([df, one_hot_df], axis=1)
            return df
        return encode
    
class ListOnehotEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        config = copy.deepcopy(self.config)
        def encode(df):
            from sklearn.preprocessing import MultiLabelBinarizer
            
            for feature, (sep, keys) in config.items():
                splitted_s = df[feature].str.split(sep).apply(lambda x: [i for i in x if i is not ""])
                mlb = MultiLabelBinarizer()
                encoded = mlb.fit_transform(splitted_s)
                names = [f"{feature}_{key}" for key in mlb.classes_]
                selected_columns = [f"{feature}_{key}" for key in keys]
                one_hot_df = pd.DataFrame(encoded, columns=names)
                one_hot_df = one_hot_df.loc[:, one_hot_df.columns.isin(selected_columns)]
                # ret = {}
                # len_df = one_hot_df.shape[0]
                # for col_name in selected_columns:
                #     if col_name in one_hot_df.columns:
                #         ret[col_name] = one_hot_df[col_name].to_numpy()
                #     else:
                #         ret[col_name] = np.zeros(len_df, dtype=np.uint8)
                # one_hot_df = pd.DataFrame(ret, dtype=np.uint8, index=df.index)
                df = pd.concat([df, one_hot_df], axis=1)
            return df
        return encode
    

class TargetEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        feature_in = copy.deepcopy(self.feature_in)
        def encode(df):
            for feature in feature_in:
                codes, uniques = pd.factorize(df[feature])
                df[f"{feature}__idx"] = pd.Series(codes, df[feature].index)
            return df
        return encode


import numpy as np
import time
import pandas as pd            
import math
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD as SparkRDD

class SparkDataFrameToRDDConverter:
    @staticmethod
    def get_function(rdp):
        print("append SparkDataFrameToRDDConverter")
        def convert(df):
            def transform(rdd_):
                rows = (row_.asDict() for row_ in rdd_)
                pdf = pd.DataFrame(rows)
                return [pdf]
            return df.rdd.mapPartitions(transform)            
        return convert
   
class RDDToSparkDataFrameConverter:
    @staticmethod
    def get_function(rdp):
        print("append RDDToSparkDataFrameConverter")
        def convert(rdd):
            df = rdd.flatMap(lambda pdf: pdf.to_dict('records')).toDF()
            return df
            
        return convert

class DataFrameToRDDConverter:
    @staticmethod
    def get_function(rdp):
        print("append DataFrameToRDDConverter")
        def convert(df):
            # convert pandas to spark
            N_PARTITIONS = 200
            num_rows = df.shape[0]
            start_time = time.time()
            permuted_indices = np.array(list(range(num_rows)))
            dfs = []
            steps = math.ceil(num_rows / N_PARTITIONS)
            end = 0
            start = 0
            while end < num_rows:
                end = start + steps if start + steps < num_rows else num_rows
                rows = permuted_indices[start : end]
                start += steps
                dfs.append(df.iloc[rows])
            rdds = rdp.spark.sparkContext.parallelize(dfs)
            end_time = time.time()
            print(f"DataframeConvert partition pandas dataframe to spark RDD took {(end_time - start_time):.3f} secs")
            return rdds
            
        return convert

class RDDToDataFrameConverter:
    @staticmethod
    def get_function(rdp):
        print("append RDDToDataFrameConverter")
        def convert(rdd):
            start_time = time.time()
            pdfs = rdd.collect()
            end_time = time.time()
            # id_pdfs = [tpl[1] for tpl in pdfs]
            # nr_pdfs = [tpl[0].shape[0] for tpl in pdfs]
            # ordered_pdfs = [tpl[0] for tpl in sorted(pdfs, key = lambda x: x[1])]
            nr_pdfs = [tpl.shape[0] for tpl in pdfs]
            ordered_pdfs = pdfs
            print(f"DataframeTransform took {(end_time - start_time):.3f} secs, processed {sum(nr_pdfs)} rows with num_partitions as {len(pdfs)}")
            start_time = time.time()
            combined = pd.concat(ordered_pdfs)
            end_time = time.time()
            print(f"DataframeTransform combine to one pandas dataframe took {(end_time - start_time):.3f} secs")
            return combined
            
        return convert

class DataFrameToSparkDataFrameConverter:
    def get_function(rdp):
        print("append DataFrameToSparkDataFrameConverter")
        def convert(pdf):
            return rdp.spark.createDataFrame(pdf) 
        return convert
    
class SparkDataFrameToDataFrameConverter:
    @staticmethod
    def get_function(rdp):
        print("append SparkDataFrameToDataFrameConverter")
        def convert(df):
            return df.to_pandas()          
        return convert


import os, requests
from tqdm import tqdm
import shutil
from pathlib import Path
import tarfile, gzip, zipfile

libpath = str(Path(__file__).parent.resolve())

class base_api:
    def __init__(self):        
        self.cache_dir = f"{libpath}/dataset_cache"
        os.makedirs(self.cache_dir, exist_ok = True)
       
    # def download_s3(self, bucket, filename):
    #     to_save = f"{self.cache_dir}/{filename}"
    #     # check if miniconda exsists
    #     if os.path.exists(to_save):
    #         return to_save
    #     s3r = boto3.resource('s3', aws_access_key_id=a1.replace("\\", ""),
    #         aws_secret_access_key=a2.replace("\\", ""))
    #     buck = s3r.Bucket(bucket)
    #     buck.download_file(filename, to_save)
    #     return to_save
        
    def get_file_path(self):
        return self.saved_path
     
    def download_url(self, name, url, unzip = False):
        to_download = url
        final_name = name
        cur_folder = self.__class__.__name__
        # skip when file exists
        if name == "":
            if os.path.exists(f"{self.cache_dir}/{cur_folder}"):
                return f"{self.cache_dir}/{cur_folder}"
        else:
            if os.path.exists(f"{self.cache_dir}/{name}"):
                return f"{self.cache_dir}/{name}"
            if os.path.exists(f"{self.cache_dir}/{cur_folder}/{name}"):
                return f"{self.cache_dir}/{cur_folder}/{name}"
        # not exists, start to download
        if unzip:
            name = "download_tmp"
        to_save = f"{self.cache_dir}/{name}"
        # check if miniconda exsists
        if not os.path.exists(to_save):
            with requests.get(to_download, stream=True) as r:
                # check header to get content length, in bytes
                total_length = int(r.headers.get("Content-Length"))
                
                # implement progress bar via tqdm
                with tqdm.wrapattr(r.raw, "read", total=total_length, desc="")as raw: 
                    # save the output to a file
                    with open(f"{to_save}", 'wb')as output:
                        shutil.copyfileobj(raw, output)   

        def unzip_tgz(src, dst):
            print("try unzip using tarfile")
            if not os.path.exists(dst):
                os.mkdir(dst)
            try:
                with tarfile.open(src) as tf:
                    tf.extractall(dst)
            except Exception as e:
                shutil.rmtree(dst)
                raise e
            
        def unzip_zip(src, dst):
            print("try unzip using zipfile")
            if not os.path.exists(dst):
                os.mkdir(dst)
            try:
                with zipfile.ZipFile(src, "r") as tf:
                    tf.extractall(dst)
            except Exception as e:
                shutil.rmtree(dst)
                raise e
        
        def unzip_gz(src, dst):
            print("try unzip using gzip")
            with gzip.open(src, 'rb') as f_in:
                with open(dst, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            
        if unzip:
            name = final_name
            unzip_success = True
            print(f"Start to unzip {to_save}")
            try:
                unzip_tgz(to_save, f"{self.cache_dir}/{cur_folder}")
                to_delete = to_save
                to_save = f"{self.cache_dir}/{cur_folder}/{name}"
            except:
                try:
                    unzip_gz(to_save, f"{self.cache_dir}/{name}")
                    to_delete = to_save
                    to_save = f"{self.cache_dir}/{name}"
                except:
                    try:
                        unzip_zip(to_save, f"{self.cache_dir}/{cur_folder}")
                        to_delete = to_save
                        to_save = f"{self.cache_dir}/{cur_folder}/{name}"
                    except:
                        print(f"downloaded, but unable to unzip {url}, please unzip manully, data saved at {to_save}")
                        unzip_success = False
            if unzip_success and os.path.exists(to_delete):
                os.remove(to_delete)
        return to_save



import unittest
import sys
import pandas as pd
from pathlib import Path
pathlib = str(Path(__file__).parent.parent.resolve())
print(pathlib)
try:
    import pyrecdp
except:
    print("Not detect system installed pyrecdp, using local one")
    sys.path.append(pathlib)
from pyrecdp.autofe import FeatureWrangler, DataEstimator
#from IPython.display import display


class TestDataEstimatorPandasBased(unittest.TestCase):
        
    def test_frauddetect_train(self):
        from pyrecdp.datasets import ibm_fraud_detect
        train_data = ibm_fraud_detect().to_pandas('test')
        data_pipeline = FeatureWrangler(dataset=train_data, label="Is Fraud?")
        def train_test_splitter(df):
            test_sample = df[df['Year'] == 2018]
            train_sample = df[df['Year'] < 2018]
            return train_sample, test_sample

        config = {
            'model_file': 'test_frauddetect.mdl',
            'metrics': 'auc', 
            'objective': 'binary', 
            'model_name': 'lightgbm',
            'train_test_splitter': train_test_splitter}
        train_pipeline = DataEstimator(method = 'train', data_pipeline = data_pipeline, config = config)
        train_pipeline.fit_transform()
        train_pipeline.export("test_fraud_detect_pipeline.json")
        
    def test_frauddetect_test(self):
        from pyrecdp.datasets import ibm_fraud_detect
        test_data = ibm_fraud_detect().to_pandas('test')
        config = {'dataset': test_data, 'label': 'Is Fraud?'}
        predict_pipeline = DataEstimator(data_pipeline = "test_fraud_detect_pipeline.json", method = 'predict', config = config)
        predict_pipeline.fit_transform()
        
    def test_frauddetect_train_with_json(self):
        from pyrecdp.datasets import ibm_fraud_detect
        test_data = ibm_fraud_detect().to_pandas('test')
        config = {'dataset': test_data, 'label': 'Is Fraud?'}
        train_pipeline = DataEstimator(data_pipeline = "test_fraud_detect_pipeline.json", method = 'train', config = config)
        train_pipeline.fit_transform()
        
    def test_frauddetect_train_with_transformed(self):
        from pyrecdp.datasets import ibm_fraud_detect
        train_data = ibm_fraud_detect().to_pandas('test')
        data_pipeline = FeatureWrangler(dataset=train_data, label="Is Fraud?")
        transformed_data = data_pipeline.fit_transform()
        #display(transformed_data.dtypes)
        def train_test_splitter(df):
            test_sample = df[df['Year'] == 2018]
            train_sample = df[df['Year'] < 2018]
            return train_sample, test_sample

        config = {
            'model_file': 'test_frauddetect.mdl',
            'metrics': 'auc', 
            'objective': 'binary', 
            'model_name': 'lightgbm',
            'train_test_splitter': train_test_splitter}
        train_pipeline = DataEstimator(method = 'train', data_pipeline = data_pipeline, config = config)
        train_pipeline.fit_transform()
        train_pipeline.export("test_fraud_detect_pipeline.json")


from .base import BaseOperation
import copy

class CustomOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.custom_op = op_base.config["func_name"]
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        return self.custom_op


from collections import defaultdict
                
class Graph:
    # Constructor
    def __init__(self):
        self.graph = defaultdict(list)
        self.nodes = []
        self.src_list = []
        self.dst_list = []
 
    def addEdge(self,u,v):
        self.graph[v].append(u)
        if u not in self.nodes:
            self.nodes.append(u)
        if v not in self.nodes:
            self.nodes.append(v)
        if v not in self.src_list:
            self.src_list.append(v)
        if u not in self.dst_list:
            self.dst_list.append(u)

    def find_last_node(self):
        last_node_candidates = []
        for n in self.nodes:
            if n in self.src_list and n not in self.dst_list:
                last_node_candidates.append(n)
        if len(last_node_candidates) == 1:
            return last_node_candidates[-1]
        elif len(last_node_candidates) == 0:
            raise ValueError(f"Found no leaf node in pipeline DAG")
        else:
            raise NotImplementedError(f"Found more than one node as leaf {last_node_candidates}, not supported yet")
        
    def bfs(self, visited, node, queue, ret): #function for BFS
        visited.add(node)
        queue.append(node)

        while queue:
            m = queue.pop(0)
            ret.append(m)

            for neighbour in self.graph[m]:
                if neighbour not in visited:
                    visited.add(neighbour)
                    queue.append(neighbour)
        return ret

    def dfs(self, visited, node, ret):  #function for dfs 
        if node not in visited:
            ret.append(node)
            visited.add(node)
            for neighbour in self.graph[node]:
                ret = self.dfs(visited, neighbour, ret)
        return ret

    def chain(self, method = 'dfs'):
        last_node = self.find_last_node()
        ret = []
        if method == 'bfs':
            # Go BFS
            queue = []
            visited = set()
            ret = self.bfs(visited, last_node, queue, ret)
        if method == 'dfs':
            # Go DFS
            visited = set()
            ret = self.dfs(visited, last_node, ret)
        ret.reverse()
        return ret
        

class DiGraph(dict):
    def get_max_idx(self):
        return max(list(self.keys()))
    
    def convert_to_node_chain(self):
        graph = Graph()
        edges = []

        for node_id, config in self.items():
            if config.children:
                for src_id in config.children:
                    edges.append([src_id, node_id])
        for edge in edges:
            graph.addEdge(*edge)
            
        ret = graph.chain()
        return ret

    def json_dump(self):
        import json
        to_dump = dict((node_id, op.dump()) for node_id, op in self.items())
        return json.dumps(to_dump, indent=4)
        
    


from pyrecdp.primitives.operations.base import BaseOperation
import numpy as np
from sklearn.metrics import mean_squared_error, roc_auc_score
from pyrecdp.core.utils import callable_string_fix

class BaseEstimator(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = self.op.config
        if 'method' in self.config:
            self.method = self.config['method']
        else:
            self.method = 'predict'

    def get_func_train(self):
        raise NotImplementedError("BaseEstimator is an abstract class")

    def get_func_predict(self):
        raise NotImplementedError("BaseEstimator is an abstract class")

    def get_function_pd(self):
        if self.method == 'train':
            return self.get_func_train()
        else:
            return self.get_func_predict()
        
    def get_evaluate_func(self, metric):
        if metric == 'rmse':
            def rmse_func(ground_truth, pred):
                return np.sqrt(mean_squared_error(ground_truth, pred))
            return rmse_func
        if metric == 'auc':
            def auc_func(ground_truth, pred):
                return roc_auc_score(ground_truth, pred)
            return auc_func

    def get_splitter_func(self, splitter_func_str):
        if splitter_func_str is not None:
            if callable(splitter_func_str):
                return splitter_func_str
            elif isinstance(splitter_func_str, str):
                splitter_func_str = callable_string_fix(splitter_func_str)
                import re
                func_name = ''.join(re.findall('def (\S+)\(', splitter_func_str.split('\n')[0]))
                print(f"Detect splitter provided, function name is {func_name}")
                print(splitter_func_str)
                exec(splitter_func_str, globals())
                return eval(func_name)
            else:
                raise NotImplementedError(f"Unable to inteprete {splitter_func_str}as train_test_splitter")
        else:
            def splitter_func(df):
                test_sample = df.sample(frac = 0.05)
                train_sample = df.drop(test_sample.index)
                return train_sample, test_sample
            return splitter_func


from .base import BaseOperation
import pandas as pd

class DataFrameOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
        self.fast_without_dpp = True

    def set(self, dataset):
        self.cache = dataset[self.op.config]
        
    def get_function_pd(self):
        cache = self.cache.copy() if self.cache is not None else None
        def get_dataframe(df):
            if df is not None:
                return df
            else:
                return cache
        return get_dataframe
    
class DataLoader(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.support_spark_dataframe = True
        self.support_spark_rdd = False
        
    def get_function_pd(self):
        def get_dataframe():
            file_path = self.op.config['file_path']
            if file_path.endswith('.csv'):
                return pd.read_csv(file_path)
            elif file_path.endswith('.parquet'):
                return pd.read_parquet(file_path)
            else:
                raise NotImplementedError("now sample read only support csv and parquet")
        return get_dataframe
    
    def get_function_spark(self, rdp, method = None):
        def get_dataframe():
            file_path = self.op.config['file_path']
            if file_path.endswith('.csv'):
                return rdp.spark.read.csv(file_path, header=True, inferSchema=True)
            elif file_path.endswith('.parquet'):
                return rdp.spark.read.parquet(file_path)
            else:
                raise NotImplementedError("now sample read only support csv and parquet")
        return get_dataframe


from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from pyrecdp.primitives.operations import Operation
from featuretools.primitives.base import TransformPrimitive

class StringToNumber(TransformPrimitive):
    name = "string_to_number"
    return_type = int

    def __init__(self):
        pass

    def get_function(self):
        def string_to_number(array):
            import re
            import pandas as pd
            array = array.apply(lambda x: "".join(re.findall(r'\d+.*\d*', x)))
            array = pd.to_numeric(array, errors='coerce')
            return array

        return string_to_number
    
class ConvertToNumberFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.op_list = [
            StringToNumber
        ]
        self.op_name = 'string_to_number'

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_re_numeric:
                in_feat_name = pa_field.name
                self.feature_in.append(in_feat_name)
        return super().fit_prepare(pipeline, children, max_idx)


import unittest
import sys
import pandas as pd
from pathlib import Path
pathlib = str(Path(__file__).parent.parent.resolve())
print(pathlib)
try:
    import pyrecdp
except:
    print("Not detect system installed pyrecdp, using local one")
    sys.path.append(pathlib)
from pyrecdp.autofe import FeatureWrangler, BasePipeline

cur_dir = str(Path(__file__).parent.resolve())

class TestPipielineJson(unittest.TestCase):

    def test_import_nyc(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_nyc_taxi_fare.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="fare_amount")

        # export pipeline to json
        pipeline.export(f"{cur_dir}/nyc_taxi_fare_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="fare_amount")
        new_pipeline.import_from_json(f"{cur_dir}/nyc_taxi_fare_pipeline.json")
        new_pipeline.fit_transform()
        
    def test_import_amazon(self):
        train_data = pd.read_table(f"{pathlib}/tests/data/amazon_reviews_us_Books.tsv", on_bad_lines='skip')
        pipeline = FeatureWrangler(dataset=train_data, label="star_rating")
        
        # export pipeline to json
        pipeline.export(f"{cur_dir}/amazon_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="star_rating")
        new_pipeline.import_from_json(f"{cur_dir}/amazon_pipeline.json")
        new_pipeline.fit_transform()
        
    def test_import_twitter(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_twitter_recsys.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="reply")
        
        # export pipeline to json
        pipeline.export(f"{cur_dir}/twitter_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="reply")
        new_pipeline.import_from_json(f"{cur_dir}/twitter_pipeline.json")
        new_pipeline.fit_transform()

    def test_import_nyc_execute_spark(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_nyc_taxi_fare.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="fare_amount")

        # export pipeline to json
        pipeline.export(f"{cur_dir}/nyc_taxi_fare_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="fare_amount")
        new_pipeline.import_from_json(f"{cur_dir}/nyc_taxi_fare_pipeline.json")
        new_pipeline.fit_transform('spark')
        

    def test_import_amazon_execute_spark(self):
        train_data = pd.read_table(f"{pathlib}/tests/data/amazon_reviews_us_Books.tsv", on_bad_lines='skip')
        pipeline = FeatureWrangler(dataset=train_data, label="star_rating")
        
        # export pipeline to json
        pipeline.export(f"{cur_dir}/amazon_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="star_rating")
        new_pipeline.import_from_json(f"{cur_dir}/amazon_pipeline.json")
        new_pipeline.fit_transform('spark')
        
    def test_import_twitter_execute_spark(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_twitter_recsys.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="reply")
        
        # export pipeline to json
        pipeline.export(f"{cur_dir}/twitter_pipeline.json")

        # create a brand-new pipeline and import from json
        new_pipeline = BasePipeline(dataset=train_data, label="reply")
        new_pipeline.import_from_json(f"{cur_dir}/twitter_pipeline.json")
        new_pipeline.fit_transform('spark')



import setuptools
from setuptools import find_packages
from setuptools.command.install import install

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name="pyrecdp",
    version="1.0.1b202304251",
    author="INTEL AIA",
    author_email="chendi.xue@intel.com",
    description=
    "A data processing bundle for spark based recommender system operations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = "https://github.com/intel/e2eAIOK/",
    project_urls={
        "Bug Tracker": "https://github.com/intel/e2eAIOK/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_dir={},
    packages=find_packages(),
    package_data={"": ["*.jar"]},
    python_requires=">=3.6",
    #cmdclass={'install': post_install},
    zip_safe=False,
    install_requires=[
        "scikit-learn",
        "psutil",
        "tqdm",
        "pyyaml",
        "pandas",
        "numpy",
        "pyarrow",
        "pandas_flavor",
        "featuretools",
        "bokeh>=2.4.2",
        "transformers",
        "ipywidgets",
        "plotly",
        "shapely",
        "graphviz",
        "requests",
        "distro",
        "pyspark==3.3.1",
        "lightgbm",
        "jupyter",
        "docx2txt",
        "openpyxl"
        ])



from .base import BaseOperation
import pandas as pd

class MergeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = op_base.config
        
    def execute_pd(self, pipeline):
        if self.cache is not None:
            return
        
        if len(self.op.children) != 2:
            raise ValueError("merge operation only accept num_children as 2")
        left_child = pipeline[self.op.children[0]].cache
        right_child = pipeline[self.op.children[1]].cache
        self.cache = pd.merge(left_child, right_child, on = self.config['on'], how = self.config['how'])

    def execute_spark(self, pipeline, rdp):
        if self.cache is not None:
            return
        
        if len(self.op.children) != 2:
            raise ValueError("merge operation only accept num_children as 2")
        
        left_child = pipeline[self.op.children[0]].cache
        right_child = pipeline[self.op.children[1]].cache
    
        self.cache = left_child.join(right_child, on = self.config['on'], how = self.config['how'])


import ipywidgets as widgets
from IPython.display import display
from pyrecdp.widgets import *
from pyrecdp.widgets.BaseWidget import *
from pyrecdp.widgets.TabWidget import *
from pyrecdp.widgets.TableViewWidget import *


from .base import BaseFeatureGenerator as super_class
from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
import pandas as pd
import numpy as np
from pandas.api import types as pdt
import copy
from featuretools.primitives.base import TransformPrimitive

def try_category(s):
    if pdt.is_categorical_dtype(s) and not pdt.is_bool_dtype(s):
        return False
    n_unique = s.nunique()
    total_len = len(s)
    threshold = (total_len / 5)
    if 1 <= n_unique <= threshold:
        return True     
    return False

def try_onehot(s):
    n_unique = s.nunique()
    if n_unique < 10:
        return s.unique().tolist()
    return False

def try_datetime(s):
    if pdt.is_datetime64_any_dtype(s):
        return False
    if not pdt.is_string_dtype(s):
        return False
    if s.isnull().all():
        return False
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        result = pd.to_datetime(result, errors='coerce', infer_datetime_format=True)
        if result.isnull().mean() > 0.8:  # If over 80% of the rows are NaN
            return False
        else:
            return True
    except:
        return False

def try_re_numeric(s):
    if s.isnull().all():
        return True
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        import re
        result = result.apply(lambda x: "".join(re.findall(r'\d+.*\d*', x)))
        result = pd.to_numeric(result, errors='coerce')
        if result.isnull().mean() <= 0.8:  # If over 80% of the rows are NaN
            return True
        return False
    except:
        return False
     
def try_numeric(s):
    if pdt.is_numeric_dtype(s):
        return True
    if s.isnull().all():
        return True
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        result = pd.to_numeric(result, errors='coerce')
        if result.isnull().mean() <= 0.8:  # If over 80% of the rows are NaN
            return True
        return False
    except:
        return False

def try_list_string(s):
    if not pdt.is_string_dtype(s):
        return False
    result = s[s.notna()]
    try:
        result = pd.Series(s).str.split(';')
        avg_words = result.str.len().mean()
        if avg_words > 1:
            estimated_size = result.explode().nunique()
            if estimated_size < 100:
                return (';', result.explode().unique().tolist())
    except:
        pass
    try:
        result = pd.Series(s).str.split(',')
        avg_words = result.str.len().mean()
        if avg_words > 1:
            estimated_size = result.explode().nunique()
            if estimated_size < 100:
                return (',', result.explode().unique().tolist())
    except:
        pass
    
    return False
         
class TypeInferFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx, df):
        ret_pa_fields = []
        pa_schema = pipeline[children[0]].output
        for pa_field, feature_name in zip(pa_schema, df.columns):
            config = {}
            if try_datetime(df[feature_name]):
                config['is_datetime'] = True            
            elif try_numeric(df[feature_name]):
                config['is_numeric'] = True
            elif try_re_numeric(df[feature_name]):
                config['is_re_numeric'] = True
            elif try_category(df[feature_name]):
                config['is_categorical'] = True
            config['is_list_string'] = try_list_string(df[feature_name])
            if config['is_list_string'] is False:
                skip = ('is_numeric' in config and config['is_numeric']) or ('is_re_numeric' in config and config['is_re_numeric'])
                if not skip:
                    config['is_onehot'] = try_onehot(df[feature_name])
            
            pa_field.copy_config_from(config)
            ret_pa_fields.append(pa_field)
        
        # append to pipeline
        cur_idx = max_idx + 1
        config = [x.mydump() for x in ret_pa_fields]
        pipeline[cur_idx] = Operation(cur_idx, children, ret_pa_fields, op = 'type_infer', config = config)
        return pipeline, cur_idx, cur_idx
    

 
class TypeCheckFeatureGenerator(super_class):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        config = {}
        for idx in range(len(pa_schema)):
            pa_field = pa_schema[idx]
            if pa_field.is_categorical_and_string:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.StringDtype())
                config[pa_field.name] = 'str'
            elif pa_field.is_categorical:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.Int32Dtype())
                config[pa_field.name] = 'int'
        
        cur_idx = max_idx
        #pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'type_check', config = config)
        return pipeline, cur_idx, cur_idx

class IntTransformer(TransformPrimitive):
    name = "astype_int"
    return_type = int

    def get_function(self):
        def astype_int(array):
            try:
                ret = array.astype(int)
            except:
                ret = array
            return ret
        return astype_int

class FloatTransformer(TransformPrimitive):
    name = "astype_float"
    return_type = float

    def get_function(self):
        def astype_float(array):
            try:
                ret = array.astype(float)
            except:
                ret = array
            return ret
        return astype_float

class TypeConvertFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        feature_in = {}
        feature_in_out_map = {}
        is_useful = False
        for pa_field in pa_schema:
            if pa_field.is_integer:
                op_class = IntTransformer
                feature_in[pa_field.name] = op_class
            if pa_field.is_float:
                op_class = FloatTransformer
                feature_in[pa_field.name] = op_class

        for in_feat_name, op_clz in feature_in.items():
            is_useful = True
            feature_in_out_map[in_feat_name] = []
            feature_in_out_map[in_feat_name].append((in_feat_name, op_clz))
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = "astype", config = feature_in_out_map)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx



from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD
import copy
 
class DropOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in = op_base.config
        self.support_spark_dataframe = True
        self.support_spark_rdd = True
        self.fast_without_dpp = True

    def get_function_pd(self):
        feature_in = copy.deepcopy(self.feature_in)
        def drop_useless_feature(df):
            return df.drop(columns = feature_in)
        return drop_useless_feature
    
    def get_function_spark(self, rdp):
        def drop_feature(df):
            return df.drop(*self.feature_in)
        return drop_feature


from pyrecdp.primitives.generators import *
from .BasePipeline import BasePipeline
from pyrecdp.core import DataFrameSchema
from pyrecdp.core.utils import sample_read
from pyrecdp.primitives.operations import Operation
import logging

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class RelationalBuilder(BasePipeline):
    def __init__(self, dataset, label, *args, **kwargs):
        super().__init__(dataset, label)

        # If we provided multiple datasets in this workload
        self.generators.append([cls() for cls in relation_builder_list])
        idx = 1
        self.children = [0]
        for table_name, table in self.supplementary.items():
            if isinstance(table, str):
                data = sample_read(table)
                self.pipeline[idx] = Operation(
                    idx, None, output = DataFrameSchema(data), op = 'DataLoader', config = {'table_name': table_name, 'file_path': table})
            else:
                self.pipeline[idx] = Operation(
                    idx, None, output = DataFrameSchema(table), op = 'DataFrame', config = table_name)
            self.children.append(idx)
            idx += 1
        self.fit_analyze()

    def fit_analyze(self, *args, **kwargs):
        child = self.children
        max_id = max(self.children)
        for i in range(len(self.generators)):
            for generator in self.generators[i]:
                child = child if isinstance(child, list) else [child]
                self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, child, max_id)


import ipywidgets as widgets
from IPython.display import display

class TabWidget:
    
    def __init__(self, children = None, display_flag = True):
        self.view = widgets.Tab()
        self.children = []
        if children:
            [self.append(c[0], c[1]) for c in children]
        if display_flag:
            display(self.view)

    def append(self, tab_name, tab_widget):
        if not self.get_tab_from_children(tab_name):
            idx = len(self.children)
            self.children.append((tab_name, tab_widget))
            self.view.children = [c[1].view for c in self.children]
            self.view.set_title(idx, tab_name)
            
    def get_tab_from_children(self, tab_name):
        for n in self.children:
            if tab_name == n[0]:
                return n
        return None



from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from featuretools.primitives import (
    Day,
    Month,
    Weekday,
    Year,
    Hour,
    PartOfDay
)
import numpy as np
import pandas as pd
from featuretools.primitives.base import TransformPrimitive
from pyrecdp.primitives.operations import Operation

class DatetimeTransformer(TransformPrimitive):
    name = "astype_datetime"
    return_type = np.datetime64

    def get_function(self):
        def convert(array):
            return pd.to_datetime(array, errors='coerce', infer_datetime_format=True)

        return convert

class DatetimeFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.op_list = [
            Day,
            Month,
            Weekday,
            Year,
            Hour,
            #PartOfDay()
        ]
        self.op_name = 'datetime_feature'            

    def fit_prepare(self, pipeline, children, max_idx):
        cur_idx = max_idx
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_datetime:
                self.feature_in.append(pa_field.name)
        feature_in_out_map = {}     
        for in_feat_name in self.feature_in:
            feature_in_out_map[in_feat_name] = []
            feature_in_out_map[in_feat_name].append((in_feat_name, DatetimeTransformer))
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = "astype", config = feature_in_out_map)
        return super().fit_prepare(pipeline, [cur_idx], cur_idx)
        


from pyrecdp.primitives.generators import *
from .BasePipeline import BasePipeline
import logging
from pyrecdp.primitives.operations import Operation
from pyrecdp.core import DiGraph
import copy
from pyrecdp.core import SparkDataProcessor

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class DataEstimator(BasePipeline):
    def __init__(self, data_pipeline, method, config = None):
        """
        Parameters: data_pipeline, method and config
        config = {
            'model_file': string,
            'dataset': path or dataframe,
            'label': string,
            'model_name': string,
            'objective': string,
            'metrics': string,
            'train_test_splitter': function
        }
        Two option:
        data_pipeline is json_file_path, specify config for ['dataset', 'label']
        data_pipeline is pre created pipeline object, specify config for ['model_name', 'objective', 'metrics']
        """
        model_file = config['model_file'] if 'model_file' in config else None
        label = config['label'] if 'label' in config else None
        self.transformed_cache = None
        if isinstance(data_pipeline, str):
            dataset = config['dataset'] 
            if dataset is not None:
                super().__init__(dataset, label)
                self.import_from_json(data_pipeline)
        elif isinstance(data_pipeline, BasePipeline):
            self.nested_pipeline_obj = data_pipeline
            self.pipeline = data_pipeline.pipeline
            self.dataset = data_pipeline.dataset
            self.rdp = data_pipeline.rdp
            self.transformed_cache = data_pipeline.transformed_cache if hasattr(data_pipeline, 'transformed_cache') else None
            label = data_pipeline.y if label is None else label
        else:
            raise NotImplementedError(f"Unsupport input datapipeline is {data_pipeline}")
        if label is None and method != 'predict':
            raise ValueError("Unable to find label for this pipeline, please provide it through API")
        
        max_idx = self.pipeline.get_max_idx()
        leaf_idx = self.pipeline.convert_to_node_chain()[-1]
        self.transformed_end_idx = leaf_idx
        
        if self.pipeline[leaf_idx].op not in ["lightgbm"]:
            model_name = config['model_name']
            objective = config['objective']
            metrics = config['metrics']
            if method == 'train':
                train_test_splitter = config['train_test_splitter'] if 'train_test_splitter' in config else None
            if model_file is None:
                model_file = f"{model_name}_{objective}_{metrics}_{label}.mdl"
            cur_idx = max_idx + 1
            self.estimator_pipeline_start = cur_idx
            # we need to add two op, one to prepare dataset, one for estimator
            op_config = {'label': label, 'metrics': metrics, 'objective': objective, 'model_file': model_file, 'method': method, 'train_test_splitter': train_test_splitter}
            op = Operation(cur_idx, [leaf_idx], None, 'DataFrame', 'main_table')
            self.pipeline[cur_idx] = op
            
            child_idx = cur_idx
            cur_idx += 1
            op_config = {'label': label, 'metrics': metrics, 'objective': objective, 'model_file': model_file, 'method': method, 'train_test_splitter': train_test_splitter}
            op = Operation(cur_idx, [child_idx], None, model_name, op_config)
            self.pipeline[cur_idx] = op
        else:
            self.pipeline[leaf_idx].config['method'] = method  

    def fit_transform(self, engine_type = 'pandas', no_cache = False, *args, **kwargs):
        if self.transformed_cache is not None and not no_cache: # we can skip data process steps
            start_op_idx = self.estimator_pipeline_start
        else:
            start_op_idx = 0
        if engine_type == "spark":
            self.rdp = SparkDataProcessor()
        ret = self.execute(engine_type, start_op_idx, no_cache, transformed_end = self.transformed_end_idx)
        if engine_type == "spark":
            del self.rdp 
            self.rdp = None
        if self.transformed_cache is None:
            self.transformed_cache = ret
        return ret    


from .base import BaseFeatureGenerator as super_class
import pandas as pd
from pyrecdp.core import SeriesSchema, DataFrameSchema
from pyrecdp.primitives.operations import Operation
from typing import List
from pyrecdp.core.utils import Timer, update_linklist

class RelationalFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pipeline = {}
        self.pipeline_start_idx = 0
        self.pipeline_main_idx = 0
        
    def fit_prepare(self, pipeline, children, max_idx):
        self.pipeline_start_idx = max_idx
        self.pipeline_main_idx = pipeline[children[0]].idx
        self.others = dict((i, pipeline[i].output) for i in children[1:])
        self.candidates = self.others.copy()
        
        pa_schema = pipeline[children[0]].output
        stop = False
        while not stop:
            related_tables = self.find_tables_share_same_feature(pa_schema)
            if len(related_tables) > 0:
                pa_schema = self.dry_run_merge_tables(pa_schema, related_tables)
            else:
                stop = True
            if len(self.candidates) == 0:
                stop = True
        pipeline.update(self.pipeline)
        return pipeline, self.pipeline_main_idx, self.pipeline_start_idx
    
    def get_function_pd(self):
        def merge_tables(df):                
            return df
        return merge_tables

    def get_function_spark(self, rdp):        
        raise NotImplementedError("RelationalFeatureGenerator spark implementation is WIP")
    
    #======== Private utility funcs ==========#
    def dry_run_merge_tables(self, target_table_schema, related_tables):            
        def _dry_merge(child_idx, on, tgt_schema, src_schema):
            tgt_col_names = [col.name for col in tgt_schema]
            src_col_names = [col.name for col in src_schema]
            error = False
            # check on is containing in both tgt and src
            for n in on:
                if n not in tgt_col_names or n not in src_col_names:
                    error = True
                    break
            if error:
                return tgt_schema
            # remove primary keys in src
            src_schema = [n for n in src_schema if n.name not in on]
            # fix conflicted col_names
            rename_cols = dict()
            for idx, n in enumerate(src_schema):
                if n.name in tgt_col_names:
                    # found conflicted name
                    fix = 1
                    still_conflict = True
                    while still_conflict:
                        cur_name = f"{n.name}_{fix}"
                        if cur_name not in tgt_col_names:
                            still_conflict = False
                        else:
                            fix += 1
                    rename_cols[n.name] = cur_name
                    src_schema[idx].name = cur_name
            # update pipeline
            left_idx = self.pipeline_main_idx
            if len(rename_cols) > 0:
                self.pipeline_start_idx += 1
                cur_idx = self.pipeline_start_idx
                self.pipeline[cur_idx] = Operation(cur_idx, [child_idx], src_schema, 'rename', rename_cols)
                right_idx = self.pipeline_start_idx
            else:
                right_idx = child_idx
            # merge to tgt
            [tgt_schema.append(n) for n in src_schema]
            
            # add merge to pipeline
            self.pipeline_start_idx += 1
            cur_idx = self.pipeline_start_idx
            self.pipeline[cur_idx] = Operation(cur_idx, [left_idx, right_idx], tgt_schema, 'merge', {'on': on, 'how': 'left'})
            self.pipeline_main_idx = cur_idx
            return tgt_schema
                
        for t_name, t_primary_keys in related_tables.items():
            target_table_schema = _dry_merge(t_name, t_primary_keys, target_table_schema, self.others[t_name])
        return target_table_schema

    def find_tables_share_same_feature(self, target_table_schema):
        related_tables = {}
        to_remove_candidates = []
        target_table_col_names = [s.name for s in target_table_schema] 
        for t, df_schema in self.candidates.items():
            df_col_names = [s.name for s in df_schema]            
            for key in df_col_names:
                if key in target_table_col_names:
                    update_linklist(related_tables, t, key)
                    to_remove_candidates.append(t)
                    
        # update candidates
        for t in to_remove_candidates:
            if t in self.candidates:
                del self.candidates[t] 
        return related_tables